# Script Name: darfortie.py
# Author: Ken Galle
# License: GPLv3
# Description: Adds functionality to the dar utility and in the process makes it
#    simpler to use, although to do this it does make some assumptions and hides
#    some functionality (which comes from the enormous list of dar switches).
#    Note however that a configuration file can be passed thru to dar.  See the
#    dar man page for details on its structure ("CONDITIONAL SYNTAX").
#    It manages the creation of incremental backups based on the most recent
#    available dar file with the same base name in a given directory.
#    It provides a way to specify a configuration file for dar without worrying
#    about the current environment (/etc/darrc, ~/.darrc).
#    It also allows you to pass thru dar prune paths.

# Note on destination naming convensions:  TODO

#    Return values:
#     1xx, where xx is any error value returned from dar during archive creation.
#     2yy, where yy is any error value returned from dar during archive testing.
#       dar return values range from 1 to 11 (see dar man page, "EXIT CODES")
#     Returns 0 on success (both create and test succeeded).
#
# TODO: attempt to uncover additional exceptions that are not handled and assign
#  error codes to these.

# TODO: test for existance of dar and if not found exit with error code.
#  Currently the subprocess fails and returns with error code of 1.

# TODO: add support for restoring a batch of archives

# TODO: packaging: http://blog.ablepear.com/2012/10/bundling-python-files-into-stand-alone.html

# TODO: Make sure dar catalog files also work as previous backup reference;
# dar has no convension for the name created (it has to be specified with -c), so we create the convension -
# "destination/asus_root_system_daily_20131227_0347UTC_catalog.1.dar"

# TODO: add option to specify path to create a catalog file of the just created archive in; defaults
# to the dest_path.

# TODO: allow the -i switch to be used for the first backup intead of erroring out.


import subprocess
import logging
import glob
import os
from darfortie import darfortie_params
import datetime
from darfortie import darfortie_previous_file

def add_dar_path_to_process_strings(params, process_strings):
    if params['dar_path'] is not None:
        process_strings.append(params['dar_path'])
    else:
        process_strings.append('dar')

def main():
    logging.basicConfig(level=logging.INFO)
    log = logging.getLogger('darfortie')
    
    # params to directly use in the call to dar
    params = {}
    
    # get params
    params = darfortie_params.parse()
    
    # build up a list of strings to send into the subprocess
    process_strings = []
    
    add_dar_path_to_process_strings(params, process_strings)
    
    # source (root) path
    process_strings.append('-R')
    process_strings.append(params['source_path'])
    
    # get the current date/time in a string usage as a suffix to the filename
    date_now = datetime.datetime.utcnow()
    date_string = date_now.strftime("%Y%m%dT%H%M") + "UTC"
    #log.info("date_string=" + date_string)
    
    # create the destination basename (includes path, if any)
    destination_basename = params['dest_path_and_base_name'] + "_" + date_string
    
    # incremental
    if params['incremental']:
        dest_basename = os.path.basename(params['dest_path_and_base_name'])
        # if user specified a different path for the previous file, swap out the paths
        if params['previous_path'] is not None:
            # get base name without path and combine with the new path
            path_and_basename_to_search = os.path.join(params['previous_path'], dest_basename)
        else:
            path_and_basename_to_search = params['dest_path_and_base_name']
        
        #log.info("path_and_basename_to_search=" + path_and_basename_to_search)
        # get list of files and dates, sort and take newest one
        # strip off .xx.dar
        full_previous_file = darfortie_previous_file.get_previous_file(
            path_and_basename_to_search, params['text_sort'])
        #log.info("full_previous_file=" + str(full_previous_file))
        if full_previous_file is not None:
            previous_file = darfortie_previous_file.remove_slice_number_and_extension(full_previous_file)
            #log.info("previous_file=" + previous_file)
            if previous_file is not None:
                process_strings.append('-A')
                process_strings.append(previous_file)
    
                # add previous date/time to current filename
                previous_datetime = darfortie_previous_file.get_previous_file_date_time(dest_basename, previous_file)
                #log.info("previous_datetime = " + previous_datetime)
                destination_basename = destination_basename + "_based_on_" + previous_datetime
            else:
                log.error("Unable to find previous file for incremental backup")
                exit(3)
        else:
            log.error("Unable to find previous file for incremental backup")
            exit(3)
    log.info("destination_basename = " + destination_basename)
    
    # archive to create
    process_strings.append('-c')
    process_strings.append(destination_basename)
    
    # define config parameter string
    # because of conditionals within config file (e.g. 'create:') this has to come after dar -c option.
    if params['config'] is not None:
        process_strings.append('--noconf')
        process_strings.append('--batch')
        process_strings.append(params['config'])
    
    # define prune paramter string
    if params['prune'] is not None:
        for onePath in params['prune']:
            process_strings.append('-P')
            process_strings.append(onePath)
    
    # exclude the destination archive(s) in case they are being created somewhere in the source directory path
    destination_name_without_path = os.path.basename(destination_basename);
    process_strings.append('-X')
    process_strings.append(destination_name_without_path + '*.*.dar');
    
    # log values from process_strings
    #for process_string in process_strings:
    #    log.info('  ' + process_string)
    
    # make call to dar
    return_code = subprocess.call(process_strings, shell=False)
    if return_code > 0:
        return_code = return_code + 100
    
    # log return value
    log.info('dar create archive return value=' + str(return_code))
    
    # if not error, then continue on to test
    if return_code == 0:
        log.info('Verifying archive...')
        test_process_strings = []
        add_dar_path_to_process_strings(params, test_process_strings)
        # test archive just created
        test_process_strings.append('-t')
        test_process_strings.append(destination_basename)
        # make call to dar
        test_return_code = subprocess.call(test_process_strings, shell=False)
        # log return value
        log.info('dar test archive return value=' + str(test_return_code))
        if test_return_code > 0:
            return_code = test_return_code + 200
    exit(return_code)

if __name__ == '__main__':
    main()



# Script Name: darfortie.py
# Author: Ken Galle
# License: GPLv3
# Description: returns a dictionary of values to be used as parameters in the dar command:

#import logging
import glob
import os.path
from operator import itemgetter

# returns previous filename
def get_previous_file(dest_path_and_base_name, text_sort):
    #log = logging.getLogger('darfortie_previous_file')

    retval = None
    #log.info("going into glob.glob = " + dest_path_and_base_name + "*.1.dar")
    filelist = glob.glob(dest_path_and_base_name + "*.1.dar")
    if len(filelist) > 0:
        filedata = []
        for onefile in filelist:
            # get the last modified datetime for each file
            filedata.append( (onefile, os.path.getmtime(onefile)) )

        # sort the list by the date
        if text_sort:
            sorted_list = sorted(filedata, key=itemgetter(0), reverse=True)
        else:
            sorted_list = sorted(filedata, key=itemgetter(1), reverse=True)

#        for myfile in sorted_list:
#            log.info("filelist item = " + myfile[0])

        retval = sorted_list[0][0]
#        log.info("chosen item = " + retval)
    return retval

def remove_slice_number_and_extension(full_previous_file):
    retval = None
    # I know the slice number will be 1 and the extension will be 'dar'
    index = full_previous_file.rfind(".1.dar", 0)
    if index > 0:
        retval = full_previous_file[0:index]
    return retval

def get_previous_file_date_time(dest_basename, previous_file):
    #log = logging.getLogger('darfortie_previous_file')

    #log.info("dest_basename = " + dest_basename)
    #log.info("previous_file = " + previous_file)
    # dest_basename: 'asus_root_system_daily'
    # previous_file:  'destination/asus_root_system_daily_20131227_0347UTC.1.dar'
    retval = None
    base_position = previous_file.rfind(dest_basename)
    if base_position is not -1:
        # calculate index of just past the base name
        date_start_index = base_position + len(dest_basename) + 1
        # the date/time will always be 16 characters long
        date_end_index = date_start_index + 16
        # slice off the date portion of the previous file
        previous_date = previous_file[date_start_index:date_end_index]
        #log.info("previous_date = " + previous_date)
        retval = previous_date
    return retval



import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "darfortie",
    version = "1.0.2",
    author = "Kenneth Galle",
    author_email = "ken.galle@rainshowers.org",
    description = "A utility that extends and simplifies using dar (Disk ARchive) for doing routine incremental backups.",
    keywords = "dar incremental backup",
    url = "https://github.com/kagalle/darfortie",
    packages=["darfortie"],
    entry_points={'console_scripts': ['darfortie = darfortie.darfortie_main:main']},
    long_description=read("README.rst"),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: BSD :: FreeBSD",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: System :: Archiving :: Backup",
    ],
)



# Script Name: darfortie.py
# Author: Ken Galle
# License: GPLv3
# Description: returns a dictionary of values to be used as parameters in the dar command:
#   dar_path : path and name to dar executable, optional, defaults to 'dar'
#   config : string, possibly None
#   prune : list of string, possibly empty
#   incremental : boolean
#   text_sort : boolean
#   source_path : string
#   dest_path_and_base_name : string

import optparse
#import logging

# Keep the dar switches explicitly on the dar command line instead of buried in variables.
# so the params{} dictionary is fine, but without putting in, e.g. -R, etc.

# provides:
#   config parameter string
#   prune parameter string
#   incremental boolean
#   text_sort boolean
def parse():
    #log = logging.getLogger('darfortie_params')
    usageString = "usage: \n%prog [common-options] [backup-options] <source_path> <dest_path_and_base_name>\n"
                  # + \
                  # "%prog [common-options] [restore-options] <dar_path_and_base_name> <destination_path>"
    descriptionString = "A front-end for dar that supports incremental backups based on " + \
    "the existing backups found in the destination folder.  <source_path> is the " + \
    "root path to back up (dar -R).  <dest_path_and_base_name> is the dar base name.  This " + \
    "may include an optional path.  This program will supply date strings to the final " + \
    "name and dar itself will supply slice numbers to form the complete filename."
    epilogString = "Based on http://dar.linux.free.fr/doc/mini-howto/dar-differential-backup-mini-howto.en.html"
    p = optparse.OptionParser(usage=usageString, description=descriptionString, epilog=epilogString)

#common options

    # -d --dar: dar filespec --> dar_path
    p.add_option("-d", "--dar", action="store", dest="dar_path", metavar="dar_filespec",
        help="filespec of dar executable; defaults to 'dar'")

    # -c --config: specify .dar config file --> conf
    p.add_option("-c", "--config", action="store", dest="conf", metavar="config_filespec",
        help="filespec of dar config file to use instead of .darrc or etc/darrc.")

#backup options

    # -P --prune: add dar prune paths --> prune
    p.add_option("-P", "--prune", action="append", dest="prune", metavar="prune_path",
        help="Specify prune paths (dar -P) to add to call to dar. Paths should be relative to " +
        "<source_path>.  This option can be repeated as needed.")

    # -i --incremental: enable incremental (dar -A) mode
    p.add_option("-i", "--incremental", action="store_true", dest="incremental", default=False,
        help="search for previous backup to use for incremental backup (dar -A).  " +
        "Finds most recent like-named backup in destination path.")

    # -I --previous_path: path to location of previous file for use by incremental option
    p.add_option("-I", "--previous_path", action="store", dest="previous_path", metavar="previous_path",
        help="alters the behavior of --incremental such that the search for a previous " +
        "backup file is done in previous_path, instead of the destination path.")

    # -i --incremental: enable incremental (dar -A) mode
    p.add_option("-t", "--text-sort", action="store_true", dest="text_sort", default=False,
        help="when searching for the latest previous backup to use, sort by file name instead " +
        "of sorting my file modification date.  For names that have yyyymmdd, etc dates/times as " +
        "part of the names.")

#restore options  (need parse_args() grouping feature

    # dictionary to return
    params = {}

    # parse command line
    opts, args = p.parse_args()

    # Note optparse errors return exit code 2.

    if len(args) != 2:
        p.print_usage()
        exit(1)

    params['dar_path'] = opts.dar_path
    params['config'] = opts.conf
    params['prune'] = opts.prune
    params['incremental'] = opts.incremental
    params['text_sort'] = opts.text_sort
    params['previous_path'] = opts.previous_path
    params['source_path'] = args[0]
    params['dest_path_and_base_name'] = args[1]

    #log.info("params:dar_path=" + str(params['dar_path']))
    #log.info("params:config=" + str(params['config']))
    #if params['prune'] is None:
    #    log.info("params:prune is None")
    #else:
    #    log.info("params:prune count=" + str(len(params['prune'])))
    #    for onePath in params['prune']:
    #        log.info("params:prune=" + str(onePath))
    #log.info("params:incremental=" + str(params['incremental']))
    #log.info("params:text_sort=" + str(params['text_sort']))
    #log.info("params:source_path=" + str(params['source_path']))
    #log.info("params:dest_path_and_base_name=" + str(params['dest_path_and_base_name']))
    #log.info("params:previous_path=" + str(params['previous_path']))

    return params
